@@ -38,6 +38,7 @@ from buildgrid.server.actioncache.caches.with_cache import WithCacheActionCache
 from buildgrid.server.actioncache.caches.write_once_cache import WriteOnceActionCache
 from buildgrid.server.actioncache.instance import ActionCache
 from buildgrid.server.bots.instance import BotsInterface
+from buildgrid.server.bots.service import UninstancedBotsService
 from buildgrid.server.build_events.storage import BuildEventStreamStorage
 from buildgrid.server.cas.instance import ByteStreamInstance, ContentAddressableStorageInstance
 from buildgrid.server.cas.storage.disk import DiskStorage
@@ -62,6 +63,7 @@ from buildgrid.server.quota.service import QuotaService
 from buildgrid.server.scheduler import (
     AgedJobHandlerOptions,
     DynamicPropertySet,
+    InstancedPropertySet,
     PropertyLabel,
     PropertySet,
     Scheduler,
@@ -823,6 +825,7 @@ def load_sql_scheduler(
     priority_assignment_percentage: int | None = None,
     job_assignment_interval: float | None = None,
     bot_poll_interval: float = 1.0,
+    proactive_fetch_to_capacity: bool = False,
 ) -> Scheduler:
     """Generates :class:`buildgrid.server.scheduler.Scheduler` using
     the tag ``!sql-scheduler``.
@@ -944,6 +947,10 @@ def load_sql_scheduler(
 
         bot_poll_interval (float): Duration to wait between bot polls for new jobs.
 
+        proactive_fetch_to_capacity (bool): When enabled, allows proactive fetching
+            up to the bot's full capacity when completing jobs, rather than just
+            fetching up to the number of completed jobs. Defaults to False.
+
         property_keys (list): Deprecated. Use a property_set instead.
         wildcard_property_keys (list): Deprecated. Use a property_set instead.
         priority_assignment_percentage (int): Deprecated. Use assigners instead.
@@ -1010,9 +1017,7 @@ def load_sql_scheduler(
             else None
         )
         queue_timeout_options = (
-            AgedJobHandlerOptions.from_config(
-                queue_timeout_job_max_age, queue_timeout_period, queue_timeout_max_window
-            )
+            AgedJobHandlerOptions.from_config(queue_timeout_job_max_age, queue_timeout_period, queue_timeout_max_window)
             if queue_timeout_job_max_age
             else None
         )
@@ -1020,7 +1025,7 @@ def load_sql_scheduler(
         sql_ro = sql_ro or sql
         sql_notifier = sql_notifier or sql
 
-        logstream_url, logstream_credentials, logstream_instance = get_logstream_connection_info(logstream)
+        logstream_url, logstream_credentials = get_logstream_connection_info(logstream)
         logstream_channel: grpc.Channel | None = None
         if logstream_url is not None:
             logstream_credentials = logstream_credentials or {}
@@ -1076,7 +1081,6 @@ def load_sql_scheduler(
             ),
             bot_session_keepalive_timeout=bot_session_keepalive_timeout,
             logstream_channel=logstream_channel,
-            logstream_instance=logstream_instance,
             asset_client=asset_client,
             queued_action_retention_hours=queued_action_retention_hours,
             completed_action_retention_hours=completed_action_retention_hours,
@@ -1089,6 +1093,7 @@ def load_sql_scheduler(
             instance_pools=instance_pools,
             bot_locality_hint_limit=bot_locality_hint_limit,
             bot_poll_interval=bot_poll_interval,
+            proactive_fetch_to_capacity=proactive_fetch_to_capacity,
         )
 
     except TypeError as type_error:
@@ -1279,6 +1284,20 @@ def load_dynamic_property_set(
     )
 
 
+_InstanceLimitedPropertySet = TypedDict("_InstanceLimitedPropertySet", {"instances": list[str], "set": PropertySet})
+
+
+@object_tag("!instanced-property-set")
+def load_instanced_property_set(sets: Iterable[_InstanceLimitedPropertySet]) -> InstancedPropertySet:
+    return InstancedPropertySet(
+        property_set_map={
+            name: instance_limited_set["set"]
+            for instance_limited_set in sets
+            for name in instance_limited_set["instances"]
+        },
+    )
+
+
 _Cohort = TypedDict("_Cohort", {"name": str, "property-labels": list[str]})
 
 
@@ -1408,6 +1427,7 @@ def load_execution_controller(
     operation_stream_keepalive_timeout: int = 600,
     endpoints: Sequence[str] = ServiceName.default_services(),
     max_list_operations_page_size: int = DEFAULT_MAX_LIST_OPERATION_PAGE_SIZE,
+    command_allowlist: list[str] | None = None,
 ) -> ExecutionController:
     """Generates :class:`buildgrid.server.execution.service.ExecutionService`
     using the tag ``!execution``.
@@ -1424,6 +1444,10 @@ def load_execution_controller(
                 - operations
                 - bots
               max-list-operations-page-size: 1000
+              command-allowlist:
+                - "gcc"
+                - "python3"
+                - "/usr/bin/make"
 
     Args:
         scheduler(:class:`Scheduler`): Instance of scheduler to use for the scheduler's state.
@@ -1436,6 +1460,10 @@ def load_execution_controller(
         max_list_operations_page_size (int): The maximum number of operations that can
             be returned in a ListOperations response. A page token will be returned
             with the response to allow the client to get the next page of results.
+        command_allowlist (list[str] | None): Optional list of allowed command binaries.
+            Commands must match exactly as specified. For example, if "gcc" is in the
+            allowlist, it will only match the exact string "gcc", not "/usr/bin/gcc".
+            If not specified, all commands are allowed.
     """
 
     click.echo(click.style(f"Creating a Execution service using {type(scheduler).__name__}\n", fg="green", bold=True))
@@ -1445,6 +1473,7 @@ def load_execution_controller(
         operation_stream_keepalive_timeout=operation_stream_keepalive_timeout,
         services=endpoints,
         max_list_operations_page_size=max_list_operations_page_size,
+        command_allowlist=command_allowlist,
     )
 
 
@@ -1535,9 +1564,7 @@ def load_with_cache_action_cache(
             code) actions. Default to ``True``.
     """
 
-    return WithCacheActionCache(
-        cache, fallback, allow_updates=allow_updates, cache_failed_actions=cache_failed_actions
-    )
+    return WithCacheActionCache(cache, fallback, allow_updates=allow_updates, cache_failed_actions=cache_failed_actions)
 
 
 @object_tag("!lru-action-cache")
@@ -2039,6 +2066,26 @@ def load_quota_service_controller(scheduler: Scheduler) -> QuotaService:
     return QuotaService(scheduler)
 
 
+@object_tag("!bots-service")
+def load_bots_service(scheduler: Scheduler) -> UninstancedBotsService:
+    """Generates :class:`buildgrid.server.bots.service.UninstancedBotsService`
+    using the tag ``!bots-service``.
+
+    Usage
+        .. code:: yaml
+
+            - !bots-service
+                scheduler: *scheduler
+
+    Args:
+        scheduler(:class:`Scheduler`): scheduler object to access job data.
+    """
+    click.echo(
+        click.style(f"Creating an uninstanced Bots service using {type(scheduler).__name__}\n", fg="green", bold=True)
+    )
+    return UninstancedBotsService(scheduler)
+
+
 @object_tag("!metering-service-client")
 def load_metering_service_client(
     base_url: str,
@@ -2102,7 +2149,6 @@ def load_metering_service_client(
 def load_asset_client(
     url: str,
     credentials: ClientCredentials | None = None,
-    instance_name: str = "",
     request_timeout: float = 5.0,
     retries: int = 3,
 ) -> AssetClient:
@@ -2114,7 +2160,6 @@ def load_asset_client(
 
             - !asset-client
               url: https://remote-asset.com
-              instance-name: dev
               credentials:
                 tls-client-cert: /path/to/cert
                 auth-token: /path/to/token
@@ -2131,7 +2176,7 @@ def load_asset_client(
         server_cert=credentials.get("tls-server-cert"),
         timeout=request_timeout,
     )
-    return AssetClient(channel=channel, instance_name=instance_name, retries=retries)
+    return AssetClient(channel=channel, retries=retries)
 
 
 @object_tag("!introspection")
@@ -2330,18 +2375,16 @@ def _validate_server_credentials(credentials: dict[str, str] | None) -> None:
         sys.exit(-1)
 
 
-def get_logstream_connection_info(logstream: Any) -> tuple[str | None, dict[str, str] | None, str | None]:
+def get_logstream_connection_info(logstream: Any) -> tuple[str | None, dict[str, str] | None]:
     logstream_url = None
     credentials = None
-    logstream_instance_name = None
     if logstream:
         logstream_url = logstream["url"]
         credentials = logstream.get("credentials")
         if not _validate_url_and_credentials(logstream_url, credentials=credentials):
             sys.exit(-1)
-        logstream_instance_name = logstream.get("instance-name", "")
 
-    return logstream_url, credentials, logstream_instance_name
+    return logstream_url, credentials
 
 
 def get_schema(strict: bool = False) -> Any:
