@@ -176,9 +176,7 @@ class ContentAddressableStorageInstance(Instance):
 
         if requested_bytes > MAX_REQUEST_SIZE:
             raise InvalidArgumentError(
-                "Combined total size of blobs exceeds "
-                "server limit. "
-                f"({requested_bytes} > {MAX_REQUEST_SIZE} [byte])"
+                f"Combined total size of blobs exceeds server limit. ({requested_bytes} > {MAX_REQUEST_SIZE} [byte])"
             )
 
         if len(good_digests) > 0:
