@@ -17,10 +17,10 @@
 SQLIndex
 ==================
 
-A SQL index implementation. This can be pointed to either a remote SQL server
-or a local SQLite database.
+A SQL index implementation. This must be pointed to a remote SQL server.
 
 """
+
 import io
 import itertools
 import time
@@ -34,6 +34,8 @@ from sqlalchemy.orm import InstrumentedAttribute, Session, load_only
 from sqlalchemy.orm.exc import StaleDataError
 from sqlalchemy.orm.query import Query
 from sqlalchemy.orm.session import Session as SessionType
+from sqlalchemy.dialects.postgresql import insert
+from sqlalchemy.sql.functions import coalesce
 
 from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
 from buildgrid._protos.google.rpc import code_pb2
@@ -48,10 +50,8 @@ from buildgrid.server.sql.provider import SqlProvider
 
 from ..storage_abc import StorageABC
 from .index_abc import IndexABC
-from .sql_dialect_delegates import PostgreSQLDelegate, SQLiteDelegate
 
 LOGGER = buildgrid_logger(__name__)
-DIALECT_DELEGATES = {"postgresql": PostgreSQLDelegate, "sqlite": SQLiteDelegate}
 
 INLINE_BLOB_SIZE_HARD_MAXIMUM = 1000000000
 
@@ -150,9 +150,6 @@ class SQLIndex(IndexABC):
             unknown_args = kwargs_keys - available_options
             raise TypeError(f"Unknown keyword arguments: [{unknown_args}]")
 
-        # Dialect-specific initialization
-        self._dialect_delegate = DIALECT_DELEGATES.get(self._sql.dialect)
-
         if inclause_limit > 0:
             if inclause_limit > window_size:
                 LOGGER.warning(
@@ -376,14 +373,32 @@ class SQLIndex(IndexABC):
 
         digest_blob_pairs = sorted(digest_blob_pairs, key=lambda pair: (pair[0].hash, pair[0].size_bytes))
 
-        if self._dialect_delegate:
-            try:
-                self._dialect_delegate._save_digests_to_index(  # type: ignore
-                    digest_blob_pairs, session, self._max_inline_blob_size
-                )
-                return
-            except AttributeError:
-                pass
+        # See discussion of __table__ typing in https://github.com/sqlalchemy/sqlalchemy/issues/9130
+        index_table = cast(Table, IndexEntry.__table__)
+        update_time = datetime.utcnow()
+        new_rows = [
+            {
+                "digest_hash": digest.hash,
+                "digest_size_bytes": digest.size_bytes,
+                "accessed_timestamp": update_time,
+                "inline_blob": (blob if digest.size_bytes <= self._max_inline_blob_size else None),
+                "deleted": False,
+            }
+            for (digest, blob) in digest_blob_pairs
+        ]
+
+        base_insert_stmt = insert(index_table).values(new_rows)
+
+        update_stmt = base_insert_stmt.on_conflict_do_update(
+            index_elements=["digest_hash"],
+            set_={
+                "accessed_timestamp": update_time,
+                "inline_blob": coalesce(base_insert_stmt.excluded.inline_blob, index_table.c.inline_blob),
+                "deleted": False,
+            },
+        )
+
+        session.execute(update_stmt)
 
         update_time = datetime.utcnow()
         # Figure out which digests we can just update
@@ -765,9 +780,7 @@ class SQLIndex(IndexABC):
 
                 if not dry_run:
                     publish_counter_metric(METRIC.STORAGE.SQL_INDEX.PREMARKED_DELETED_COUNT, len(marked_digests))
-                LOGGER.info(
-                    "Gathered pre-marked bytes.", tags=dict(collected_bytes=collected_bytes, max_bytes=n_bytes)
-                )
+                LOGGER.info("Gathered pre-marked bytes.", tags=dict(collected_bytes=collected_bytes, max_bytes=n_bytes))
                 self._delete_premarked_blobs = False
 
             if collected_bytes < n_bytes:
