@@ -207,7 +207,8 @@ class S3Storage(StorageABC):
         #    expression has type "list[dict[str, str]]",
         #    TypedDict item "Objects" has type "Sequence[ObjectIdentifierTypeDef]"
         response = self._s3.delete_objects(
-            Bucket=bucket_name, Delete={"Objects": digests}  # type: ignore[typeddict-item]
+            Bucket=bucket_name,
+            Delete={"Objects": digests},  # type: ignore[typeddict-item]
         )
         return_failed = []
         failed_deletions = response.get("Errors", [])
@@ -355,8 +356,7 @@ class S3Storage(StorageABC):
                     failed_deletions += self._multi_delete_blobs(bucket, requests[i : i + self._page_size])
                 except ClientError as error:
                     current_failed_deletions = [
-                        self._deconstruct_key(key_versions["Key"])
-                        for key_versions in requests[i : i + self._page_size]
+                        self._deconstruct_key(key_versions["Key"]) for key_versions in requests[i : i + self._page_size]
                     ]
                     failed_deletions += [
                         f"{digest_hash}/{digest_size_bytes}"
