@@ -97,7 +97,9 @@ def merkle_tree_maker(directory_path: str) -> Iterator[tuple[FileNode | Director
     directory_name = os.path.basename(directory_path)
 
     # Actual generator, yields recursively FileNodes and DirectoryNodes:
-    def __merkle_tree_maker(directory_path: str, directory_name: str) -> Generator[
+    def __merkle_tree_maker(
+        directory_path: str, directory_name: str
+    ) -> Generator[
         tuple[FileNode | DirectoryNode, BinaryIO, str],
         None,
         tuple[FileNode | DirectoryNode, BinaryIO, str],
@@ -474,9 +476,7 @@ class Downloader:
         assert len(read_blob) == digest.size_bytes
         return bytes(read_blob)
 
-    def _fetch_blob_batch(
-        self, digests: list[Digest], *, skip_unavailable: bool = False
-    ) -> list[tuple[bytes, Digest]]:
+    def _fetch_blob_batch(self, digests: list[Digest], *, skip_unavailable: bool = False) -> list[tuple[bytes, Digest]]:
         """Fetches blobs using ContentAddressableStorage.BatchReadBlobs()
         Returns (data, digest) pairs"""
 
