@@ -311,7 +311,7 @@ def _curl_multi_run(
             m.remove_handle(c)
         curl_handles.clear()
 
-    if retry_objects and attempt < S3_MAX_RETRIES:
+    if retry_objects:
         # Wait between attempts with truncated exponential backoff with jitter
         exp_backoff = 2 ** (attempt - 1)
         exp_backoff_with_jitter = random.random() * exp_backoff
