@@ -14,7 +14,7 @@
 
 from .assigner import AssignerConfig, PriorityAgeAssignerConfig
 from .impl import AgedJobHandlerOptions, Scheduler, SchedulerMetrics
-from .properties import DynamicPropertySet, PropertyLabel, PropertySet, StaticPropertySet
+from .properties import DynamicPropertySet, InstancedPropertySet, PropertyLabel, PropertySet, StaticPropertySet
 
 __all__ = [
     "AgedJobHandlerOptions",
@@ -22,6 +22,7 @@ __all__ = [
     "SchedulerMetrics",
     "PropertySet",
     "DynamicPropertySet",
+    "InstancedPropertySet",
     "StaticPropertySet",
     "PropertyLabel",
     "AssignerConfig",
