@@ -15,11 +15,10 @@
 
 from contextlib import contextmanager
 from datetime import timedelta
-from tempfile import NamedTemporaryFile
 from threading import Lock
 from typing import Any, Generator, Iterator
 
-from sqlalchemy import create_engine, event, text
+from sqlalchemy import create_engine
 from sqlalchemy.engine import Engine
 from sqlalchemy.orm import Session, scoped_session, sessionmaker
 from sqlalchemy.pool import NullPool
@@ -37,9 +36,8 @@ from buildgrid.server.sql.models import Base
 
 from .utils import (
     SQLPoolDisposeHelper,
-    is_psycopg2_connection_string,
-    is_sqlite_connection_string,
-    is_sqlite_inmemory_connection_string,
+    is_postgresql_connection_string,
+    USE_POSTGRES_MESSAGE,
 )
 
 LOGGER = buildgrid_logger(__name__)
@@ -47,15 +45,13 @@ LOGGER = buildgrid_logger(__name__)
 # Each dialect has a limit on the number of bind parameters allowed. This
 # matters because it determines how large we can allow our IN clauses to get.
 #
-# SQLite: 1000 https://www.sqlite.org/limits.html#max_variable_number
 # PostgreSQL: 32767 (Int16.MAX_VALUE) https://www.postgresql.org/docs/9.4/protocol-message-formats.html
 #
 # We'll refer to this as the "inlimit" in the code. The inlimits are
 # set to 75% of the bind parameter limit of the implementation.
-DIALECT_INLIMIT_MAP = {"postgresql": 24000, "sqlite": 750}
+DIALECT_INLIMIT_MAP = {"postgresql": 24000}
 DEFAULT_INLIMIT = 100
 
-
 # NOTE: Obviously these type annotations are useless, but sadly they're what
 # is in the upstream sqlalchemy2-stubs[0].
 #
@@ -64,24 +60,6 @@ DEFAULT_INLIMIT = 100
 #
 # [0]: https://github.com/sqlalchemy/sqlalchemy2-stubs/blob/main/sqlalchemy-stubs/pool/events.pyi#L9
 # [1]: https://github.com/sqlalchemy/sqlalchemy/blob/main/lib/sqlalchemy/pool/events.py#L96-L100
-def _sqlite_on_connect(conn: Any, record: Any) -> None:
-    """SQLite ``PRAGMA`` statements to execute immediately upon connection.
-
-    These statements configure the behaviour of the database, and are specific
-    to SQLite.
-
-    See https://www.sqlite.org/pragma.html for details.
-
-    Args:
-        conn (DBAPIConnection): The DBAPI connection that was just connected.
-        record (_ConnectionRecord): The connection record which contains the
-            DBAPI connection.
-
-    """
-    # Use journal_mode=WAL to allow read/write concurrency, as well as the
-    # performance improvements it brings.
-    conn.execute("PRAGMA journal_mode=WAL")
-    conn.execute("PRAGMA synchronous=NORMAL")
 
 
 class SqlProvider:
@@ -93,17 +71,14 @@ class SqlProvider:
 
     Args:
         connection_string (str | None): The connection string to use when
-            creating a database connection. If ``None`` then a temporary
-            SQLite database will be created for the lifetime of this
-            object.
+            creating a database connection. Must be a valid postgres database.
 
         connection_timeout (int): The timeout to use when attempting to
             connect to the database, in seconds. Defaults to 5 seconds if
             unset.
 
         lock_timeout (int): The timeout to use when the connection
-        holds a lock in the database. This is supported only if the database
-        backend is PostgresQL.
+        holds a lock in the database.
 
         connect_args (dict[str, Any] | None): Dictionary of DBAPI
             connection arguments to pass to the engine. See the
@@ -132,7 +107,7 @@ class SqlProvider:
             publishing.
 
     Raises:
-        ValueError: when ``connection_string`` specifies an in-memory SQLite
+        ValueError: when ``connection_string`` doesn't specify a Postgresql
             database.
 
     .. _docs: https://docs.sqlalchemy.org/en/14/core/engines.html#use-the-connect-args-dictionary-parameter
@@ -155,16 +130,10 @@ class SqlProvider:
     ):
         """Initialize an SqlProvider."""
         self._database_tempfile = None
-        # If we don't have a connection string, we'll make a tempfile to use
-        # as an SQLite database. This tempfile needs to exist for the lifetime
-        # of the SqlProvider.
+        # If we don't have a connection string, we'll throw a ValueError and some info about setting up a
+        # postgres database.
         if not connection_string:
-            self._database_tempfile = NamedTemporaryFile(prefix="bgd-", suffix=".db")
-            LOGGER.warning(
-                "No connection string specified for the SQL provider, will use SQLite with tempfile.",
-                tags=dict(tempfile=self._database_tempfile.name),
-            )
-            connection_string = f"sqlite:///{self._database_tempfile.name}"
+            raise ValueError(f"No connection string specified for the SQL provider\n\n{USE_POSTGRES_MESSAGE}")
 
         # Set up database connection
         self._session_factory = sessionmaker(future=True)
@@ -224,8 +193,7 @@ class SqlProvider:
                 is applied.
 
             lock_timeout (int): The timeout to use when the connection
-            holds a lock in the database. This is supported only if the database
-            backend is PostgresQL.
+            holds a lock in the database.
 
             connect_args: Dictionary of DBAPI
                 connection arguments to pass to the engine. See the
@@ -257,20 +225,24 @@ class SqlProvider:
                 database defined by ``connection_string``.
 
         Raises:
-            ValueError: when attempting to connect to an in-memory SQLite
+            ValueError: when attempting to connect to a non Postgresql
                 database.
 
         .. _docs: https://docs.sqlalchemy.org/en/14/core/engines.html#use-the-connect-args-dictionary-parameter
 
         """
+        # Disallow sqlite for the scheduler db
+        # theres no reason to support a non production ready scheduler implementation
+
         # Disallow sqlite in-memory because multi-threaded access to it is
         # complex and potentially problematic at best
         # ref: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#threading-pooling-behavior
-        if is_sqlite_inmemory_connection_string(connection_string):
+
+        # Ensure only postgres is supported
+
+        if not is_postgresql_connection_string(connection_string):
             raise ValueError(
-                "Cannot use SQLite in-memory with BuildGrid "
-                f"(connection_string=[{connection_string}]). Use a file or "
-                "leave the connection_string empty for a tempfile."
+                f"Cannot use database (connection_string=[{connection_string}]).\n\n{USE_POSTGRES_MESSAGE}"
             )
 
         extra_engine_args: dict[str, Any] = {}
@@ -280,12 +252,9 @@ class SqlProvider:
             extra_engine_args["connect_args"] = {}
 
         if connection_timeout > 0:
-            if is_sqlite_connection_string(connection_string):
-                extra_engine_args["connect_args"]["timeout"] = connection_timeout
-            elif is_psycopg2_connection_string(connection_string):
-                extra_engine_args["connect_args"]["connect_timeout"] = connection_timeout
-        if lock_timeout > 0 and is_psycopg2_connection_string(connection_string):
-            # Additional postgres specific timeouts
+            extra_engine_args["connect_args"]["connect_timeout"] = connection_timeout
+        if lock_timeout > 0:
+            # Additional timeouts
             # Additional libpg options
             # Note that those timeouts are in milliseconds (so *1000)
             # User might specifically set options... do not override in this case.
@@ -311,10 +280,6 @@ class SqlProvider:
         engine = create_engine(connection_string, echo=False, future=True, **extra_engine_args)
         self._session_factory.configure(bind=engine)
 
-        # Register sqlite-specific connection callback.
-        if engine.dialect.name == "sqlite":
-            event.listen(engine, "connect", _sqlite_on_connect)
-
         return engine
 
     @property
@@ -332,7 +297,6 @@ class SqlProvider:
         self,
         *,
         scoped: bool = False,
-        sqlite_lock_immediately: bool = False,
         exceptions_to_not_raise_on: list[type[Exception]] | None = None,
         exceptions_to_not_rollback_on: list[type[Exception]] | None = None,
         expire_on_commit: bool = True,
@@ -347,10 +311,6 @@ class SqlProvider:
         if the underlying connection pool has recently been disposed of and
         refreshed due to connectivity issues.
 
-        When ``sqlite_lock_immediately`` is ``True``, the Session will not
-        yield until the database has been locked by entering into a write
-        transaction when using SQLite.
-
         If an Exception is raised whilst in the managed context, the ongoing
         database transaction is rolled back, and the Exception is reraised.
         Some Exceptions which suggest a transient connection issue with the
@@ -371,13 +331,6 @@ class SqlProvider:
                 session. This results in reuse of the underlying Session object
                 in a given thread.
 
-            sqlite_lock_immediately: If true, execute a ``BEGIN IMMEDIATE``
-                statement as soon as the session is created when using SQLite.
-                This allows locking for the lifetime of the ``Session`` within
-                this ContextManager, enabling similar behaviour to
-                ``SELECT ... FOR UPDATE`` in other dialects. Defaults to
-                ``False``.
-
             exceptions_to_not_raise_on: The list of error types to be suppressed
                 within the context rather than re-raised. Defaults to ``None``,
                 meaning all SQLAlchemyErrors will be re-raised.
@@ -427,8 +380,6 @@ class SqlProvider:
         # Try to obtain a session
         try:
             session = factory() if scoped else factory(expire_on_commit=expire_on_commit)
-            if sqlite_lock_immediately and session.bind.name == "sqlite":  # type: ignore
-                session.execute(text("BEGIN IMMEDIATE"))
         except Exception as e:
             LOGGER.error("Unable to obtain a database session.", exc_info=True)
             raise DatabaseError("Unable to obtain a database session.") from e
@@ -474,7 +425,6 @@ class SqlProvider:
     def scoped_session(
         self,
         *,
-        sqlite_lock_immediately: bool = False,
         exceptions_to_not_raise_on: list[type[Exception]] | None = None,
         exceptions_to_not_rollback_on: list[type[Exception]] | None = None,
     ) -> Generator[Session, None, None]:
@@ -497,7 +447,6 @@ class SqlProvider:
         """
         with self.session(
             scoped=True,
-            sqlite_lock_immediately=sqlite_lock_immediately,
             exceptions_to_not_raise_on=exceptions_to_not_raise_on,
             exceptions_to_not_rollback_on=exceptions_to_not_rollback_on,
         ) as session:
