@@ -24,7 +24,12 @@ from grpc._server import _Context
 from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Action, Command, Digest
 from buildgrid._protos.google.longrunning import operations_pb2
 from buildgrid.server.enums import MeteringThrottleAction
-from buildgrid.server.exceptions import CancelledError, FailedPreconditionError, ResourceExhaustedError
+from buildgrid.server.exceptions import (
+    CancelledError,
+    FailedPreconditionError,
+    ResourceExhaustedError,
+    InvalidArgumentError,
+)
 from buildgrid.server.execution.instance import ExecutionInstance
 from buildgrid.server.scheduler import DynamicPropertySet
 from buildgrid.server.sql.models import ClientIdentityEntry
@@ -65,11 +70,16 @@ class DeprecatedAction:
 
 
 class MockStorage:
-    def __init__(self, pairs):
+    def __init__(self, pairs, command_arguments=None):
         command_digest = Digest(hash="command_digest", size_bytes=0)
         input_root_digest = Digest(hash="input_root_digest", size_bytes=0)
         self.action = Action(command_digest=command_digest, input_root_digest=input_root_digest, do_not_cache=True)
         self.command = Command()
+
+        # Add command arguments if provided
+        if command_arguments:
+            self.command.arguments.extend(command_arguments)
+
         for property_pair in pairs:
             prop = self.command.platform.properties.add()
             prop.name = property_pair[0]
@@ -151,6 +161,21 @@ def test_execute_platform_matching_too_many_os_platform_key():
         exec_instance.execute(action_digest="fake", skip_cache_lookup=False)
 
 
+def test_execute_platform_invalid_property_label():
+    """Property labels will have invalid characters and not execute"""
+
+    pairs = [("OSFamily", "linux"), ("ISA", "platform=x86-64")]
+    property_set = DynamicPropertySet(
+        unique_property_keys={"OSFamily"},
+        match_property_keys={"OSFamily", "ISA"},
+        wildcard_property_keys=set(),
+        label_key="ISA",
+    )
+    exec_instance = ExecutionInstance(MockScheduler(MockStorage(pairs), property_set=property_set))
+    with pytest.raises(InvalidArgumentError):
+        exec_instance.execute(action_digest="fake", skip_cache_lookup=False)
+
+
 def test_execute_platform_matching_failure():
     """Will not match due to platform-keys missing 'ChrootDigest'."""
 
@@ -478,3 +503,65 @@ def test_execute_throttled_rejected(
         exec_instance.execute(
             action_digest="fake", skip_cache_lookup=False, priority=-1, client_identity=client_identity
         )
+
+
+def test_execute_command_allowlist_allowed():
+    """Test that allowed commands pass validation."""
+    pairs = [("OSFamily", "linux")]
+    storage = MockStorage(pairs, command_arguments=["gcc", "-o", "test", "test.c"])
+    property_set = DynamicPropertySet(
+        unique_property_keys={"OSFamily"},
+        match_property_keys={"OSFamily"},
+        wildcard_property_keys=set(),
+    )
+    exec_instance = ExecutionInstance(
+        MockScheduler(storage, property_set=property_set), command_allowlist=["gcc", "clang"]
+    )
+    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"
+
+
+def test_execute_command_allowlist_blocked():
+    """Test that disallowed commands are rejected."""
+    pairs = [("OSFamily", "linux")]
+    storage = MockStorage(pairs, command_arguments=["python", "script.py"])
+    property_set = DynamicPropertySet(
+        unique_property_keys={"OSFamily"},
+        match_property_keys={"OSFamily"},
+        wildcard_property_keys=set(),
+    )
+    exec_instance = ExecutionInstance(
+        MockScheduler(storage, property_set=property_set), command_allowlist=["gcc", "clang"]
+    )
+    with pytest.raises(FailedPreconditionError, match="Command 'python' is not in the allowed command list"):
+        exec_instance.execute(action_digest="fake", skip_cache_lookup=False)
+
+
+def test_execute_command_allowlist_full_path():
+    """Test that full path commands require exact matching."""
+    pairs = [("OSFamily", "linux")]
+    storage = MockStorage(pairs, command_arguments=["/usr/bin/gcc", "-o", "test", "test.c"])
+    property_set = DynamicPropertySet(
+        unique_property_keys={"OSFamily"},
+        match_property_keys={"OSFamily"},
+        wildcard_property_keys=set(),
+    )
+    exec_instance = ExecutionInstance(
+        MockScheduler(storage, property_set=property_set),
+        command_allowlist=["gcc", "python"],  # Only basename "gcc", not full path
+    )
+    # Should NOT match - requires exact path match
+    with pytest.raises(FailedPreconditionError, match="Command '/usr/bin/gcc' is not in the allowed command list"):
+        exec_instance.execute(action_digest="fake", skip_cache_lookup=False)
+
+
+def test_execute_no_allowlist_allows_all():
+    """Test that no allowlist allows all commands."""
+    pairs = [("OSFamily", "linux")]
+    storage = MockStorage(pairs, command_arguments=["any_command", "--help"])
+    property_set = DynamicPropertySet(
+        unique_property_keys={"OSFamily"},
+        match_property_keys={"OSFamily"},
+        wildcard_property_keys=set(),
+    )
+    exec_instance = ExecutionInstance(MockScheduler(storage, property_set=property_set))
+    assert exec_instance.execute(action_digest="fake", skip_cache_lookup=False) == "queued"
