@@ -14,7 +14,6 @@
 
 
 import datetime
-import tempfile
 import time
 import uuid
 from concurrent.futures import ThreadPoolExecutor
@@ -76,7 +75,12 @@ from buildgrid.server.sql.models import (
 )
 from buildgrid.server.sql.provider import SqlProvider
 from buildgrid.server.utils.digests import create_digest
-from tests.utils.fixtures import mock_logstream_client, pg_schema_provider
+from tests.utils.fixtures import (  # noqa: F401 allow fixture inclusion
+    connection_string_from_connection,
+    mock_logstream_client,
+    new_postgres_fixture,
+    pg_schema_provider,
+)
 from tests.utils.server import MockAssetPushServicer
 from tests.utils.utils import find_free_port, generate_bot_name
 
@@ -119,7 +123,7 @@ def asset_service_channel(asset_service_remote: str) -> Iterator[grpc.Channel]:
 
 @pytest.fixture
 def asset_client(push_service, asset_service_channel: grpc.Channel) -> Iterator[AssetClient]:
-    yield AssetClient(asset_service_channel, instance_name="asset")
+    yield AssetClient(asset_service_channel)
 
 
 @pytest.fixture
@@ -137,15 +141,16 @@ PARAMS = ["action-cache", "no-action-cache"]
 
 @pytest.fixture(params=PARAMS)
 def controller(
-    request, asset_client: AssetClient, common_props: PropertySet, cohort_set: CohortSet
+    new_postgres, request, asset_client: AssetClient, common_props: PropertySet, cohort_set: CohortSet
 ) -> Iterator[ExecutionController]:
     use_cache = request.param
     storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
-    db = tempfile.NamedTemporaryFile().name
-    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}")
+    connection = connection_string_from_connection(new_postgres)
+    sql_provider = SqlProvider(connection_string=connection)
 
     # Create database tables
-    Base.metadata.create_all(sql_provider._engine)
+    with sql_provider._engine.begin() as conn:
+        Base.metadata.create_all(conn)
 
     with create_write_session(command_digest) as write_session:
         write_session.write(command.SerializeToString())
@@ -190,14 +195,15 @@ PARAMS_MAX_EXECUTION = [None, 0.1, 2]
 
 
 @pytest.fixture(params=PARAMS_MAX_EXECUTION)
-def controller_max_execution_timeout(request, common_props):
+def controller_max_execution_timeout(new_postgres, request, common_props):
     max_execution_timeout = request.param
     storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
-    db = tempfile.NamedTemporaryFile().name
-    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}")
+    connection = connection_string_from_connection(new_postgres)
+    sql_provider = SqlProvider(connection_string=connection)
 
     # Create database tables
-    Base.metadata.create_all(sql_provider._engine)
+    with sql_provider._engine.begin() as conn:
+        Base.metadata.create_all(conn)
 
     with create_write_session(command_digest) as write_session:
         write_session.write(command.SerializeToString())
@@ -225,14 +231,15 @@ PARAMS_MAX_QUEUE_SIZE = [None, 1, 5]
 
 
 @pytest.fixture(params=PARAMS_MAX_QUEUE_SIZE)
-def controller_max_queue_size(request, common_props):
+def controller_max_queue_size(new_postgres, request, common_props):
     max_queue_size = request.param
     storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
-    db = tempfile.NamedTemporaryFile().name
-    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}")
+    connection = connection_string_from_connection(new_postgres)
+    sql_provider = SqlProvider(connection_string=connection)
 
     # Create database tables
-    Base.metadata.create_all(sql_provider._engine)
+    with sql_provider._engine.begin() as conn:
+        Base.metadata.create_all(conn)
 
     with create_write_session(command_digest) as write_session:
         write_session.write(command.SerializeToString())
@@ -256,13 +263,14 @@ def controller_max_queue_size(request, common_props):
 
 
 @pytest.fixture
-def isolated_controllers(request, common_props):
+def isolated_controllers(new_postgres, request, common_props):
     storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
-    db = tempfile.NamedTemporaryFile().name
-    sql_provider = SqlProvider(connection_string=f"sqlite:///{db}")
+    connection = connection_string_from_connection(new_postgres)
+    sql_provider = SqlProvider(connection_string=connection)
 
     # Create database tables
-    Base.metadata.create_all(sql_provider._engine)
+    with sql_provider._engine.begin() as conn:
+        Base.metadata.create_all(conn)
 
     with create_write_session(command_digest) as write_session:
         write_session.write(command.SerializeToString())
@@ -370,7 +378,7 @@ def test_update_lease_state(controller, push_service: MockAssetPushServicer):
         # check retention for queued action
         assert job.action_digest in push_service.blobs
         asset = push_service.blobs[job.action_digest]
-        assert asset.instance_name == "asset"
+        assert asset.instance_name == "sql"
         assert asset.uris == [f"nih:sha-256;{string_to_digest(job.action_digest).hash}"]
         action = Action.FromString(job.action)
         assert action.command_digest in asset.references_blobs
@@ -399,7 +407,7 @@ def test_update_lease_state(controller, push_service: MockAssetPushServicer):
         action_result = scheduler.storage.get_message(string_to_digest(job.result), ExecuteResponse).result
         result_digest = create_digest(action_result.SerializeToString())
         result_asset = push_service.blobs[digest_to_string(result_digest)]
-        assert result_asset.instance_name == "asset"
+        assert result_asset.instance_name == "sql"
         assert result_asset.uris == [f"nih:sha-256;{result_digest.hash}"]
         assert mock_stderr_digest in result_asset.references_blobs
         assert result_asset.expire_at.ToDatetime() > datetime.datetime.now() + datetime.timedelta(hours=2)
@@ -559,7 +567,6 @@ def test_action_deduplication(controller):
 def test_get_metadata_for_leases(controller):
     scheduler = controller.execution_instance.scheduler
     scheduler.logstream_channel = mock.MagicMock()
-    scheduler.logstream_instance = "testing"
 
     with mock.patch("buildgrid.server.scheduler.impl.logstream_client", new=mock_logstream_client()):
         bot_name, bot_id, operation_name, _ = mock_queue_job_action(scheduler)
@@ -716,17 +723,27 @@ def test_get_job_operation_failed(controller):
 @pytest.mark.parametrize("n_operations", [1, 2])
 @pytest.mark.parametrize("sleep_time", [0.2])
 def test_max_execution_timeout(controller_max_execution_timeout, sleep_time, n_operations):
-    scheduler = controller_max_execution_timeout.execution_instance.scheduler
+    scheduler: Scheduler = controller_max_execution_timeout.execution_instance.scheduler
 
     max_execution_timeout = scheduler.max_execution_timeout
 
     # Queue Job
     bot_name, bot_id, operation_name, job_name = mock_queue_job_action(scheduler)
+    assert job_name is not None
 
     # Create n operations
     operation_names = [operation_name]
-    for i in range(n_operations - 1):
-        operation_names.append(scheduler.create_operation(job_name=job_name))
+    with scheduler._sql.session() as session:
+        for i in range(n_operations - 1):
+            operation_names.append(
+                scheduler._create_operation(
+                    session,
+                    job_name=job_name,
+                    client_identity=None,
+                    request_metadata=None,
+                    operation_count=n_operations,
+                )
+            )
 
     for i in range(n_operations):
         operation_initially = scheduler.load_operation(operation_names[i])
@@ -777,9 +794,7 @@ def test_max_execution_timeout_rejection(controller):
     scheduler = controller.execution_instance.scheduler
     scheduler.max_execution_timeout = 1
 
-    action = remote_execution_pb2.Action(
-        command_digest=command_digest, do_not_cache=False, timeout=Duration(seconds=2)
-    )
+    action = remote_execution_pb2.Action(command_digest=command_digest, do_not_cache=False, timeout=Duration(seconds=2))
     action_digest = create_digest(action.SerializeToString())
 
     with create_write_session(action_digest) as write_session:
@@ -905,21 +920,6 @@ def test_publish_execution_stats_no_client(controller, mock_metering_client, moc
     assert len(mock_metering_client.sink) == 0
 
 
-def test_instance_isolation_create_operation(isolated_controllers):
-    controller_a, controller_b = isolated_controllers
-    scheduler_a = controller_a.execution_instance.scheduler
-    scheduler_b = controller_b.execution_instance.scheduler
-
-    with instance_context("instance-a"):
-        _, _, _, job_name = mock_queue_job_action(scheduler_a)
-        # Check that we can only create an Operation for the job on the same instance
-        scheduler_a.create_operation(job_name=job_name)
-
-    with instance_context("instance-b"):
-        with pytest.raises(NotFoundError):
-            scheduler_b.create_operation(job_name=job_name)
-
-
 def test_instance_isolation_load_operation(isolated_controllers):
     controller_a, controller_b = isolated_controllers
     scheduler_a = controller_a.execution_instance.scheduler
@@ -1039,9 +1039,98 @@ def test_wildcard_bot_lease_assignment(isolated_controllers):
         leases = scheduler_b.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, [])
         assert len(leases) == 1
 
-        metadata = scheduler_b.get_metadata_for_leases(leases, only_current_instance=False)
+        metadata = scheduler_b.get_metadata_for_leases(leases)
         assert len(metadata) == 1
 
+        # Complete the lease
+        leases[0].state = LeaseState.COMPLETED.value
+        leases = scheduler_b.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, leases)
+        assert len(leases) == 0
+
+        with scheduler_a._sql.session() as session:
+            job = session.execute(
+                select(JobEntry).where(JobEntry.stage < OperationStage.COMPLETED.value)
+            ).scalar_one_or_none()
+            assert job is None
+
+
+def test_wildcard_bot_retries_jobs(controller):
+    scheduler: Scheduler = controller.execution_instance.scheduler
+    bot_id = "test-worker"
+
+    # Queue a job into instance-a
+    with instance_context("instance-a"):
+        _, _, _, job_name = mock_queue_job_action(scheduler, create_bot=False, assign=False)
+        assert job_name is not None
+
+    # Connect a wildcard bot
+    with instance_context("*"):
+        scheduler.add_bot_entry(
+            bot_name=f"*/{uuid.uuid4()}",
+            bot_session_id=bot_id,
+            bot_session_status=BotStatus.OK.value,
+            instance_name="*",
+        )
+
+    # Attempt to assign work in the scheduler that got the job
+    scheduler.assign_job_by_priority(failure_backoff=0)
+
+    # The job should be assigned
+    with instance_context("instance-a"):
+        with scheduler._sql.session() as session:
+            job = scheduler._get_job(job_name, session)
+            assert job is not None
+            assert job.assigned
+
+    # Reconnect a bot with the same session ID
+    with instance_context("*"):
+        scheduler.add_bot_entry(
+            bot_name=f"*/{uuid.uuid4()}",
+            bot_session_id=bot_id,
+            bot_session_status=BotStatus.OK.value,
+            instance_name="*",
+        )
+
+    # The job should have been re-queued
+    with instance_context("instance-a"):
+        with scheduler._sql.session() as session:
+            job = scheduler._get_job(job_name, session)
+            assert job is not None
+            assert not job.assigned
+
+
+def test_wildcard_bot_deletes_old_sessions(controller):
+    scheduler: Scheduler = controller.execution_instance.scheduler
+    bot_id = "test-worker"
+
+    # Connect a bot
+    with instance_context("instance-a"):
+        scheduler.add_bot_entry(
+            bot_name=f"*/{uuid.uuid4()}",
+            bot_session_id=bot_id,
+            bot_session_status=BotStatus.OK.value,
+            instance_name="instance-a",
+        )
+
+    # Reconnect a bot with the same session ID
+    with instance_context("*"):
+        scheduler.add_bot_entry(
+            bot_name=f"*/{uuid.uuid4()}",
+            bot_session_id=bot_id,
+            bot_session_status=BotStatus.OK.value,
+            instance_name="*",
+        )
+
+    # There should only be a single bot with the given ID, and it should be the one using the
+    # wildcard instance name.
+    with scheduler._sql.session() as session:
+        stmt = select(BotEntry).where(BotEntry.bot_id == bot_id)
+        bots = session.execute(stmt).scalars().all()
+        assert len(bots) == 1
+
+        bot = bots[0]
+        assert bot.instance_name == "*"
+
 
 def test_max_queue_size(controller_max_queue_size):
     scheduler = controller_max_queue_size.execution_instance.scheduler
@@ -1171,6 +1260,69 @@ def test_complete_job_with_locality_hints_limit_size(controller, controller_post
             assert len(locality_hints) == scheduler.bot_locality_hint_limit
 
 
+def test_locality_hint_not_recorded_on_non_zero_exit_code(controller_postgres):
+    """Test that locality hint is not recorded when action result has non-zero exit code."""
+    scheduler: Scheduler = controller_postgres.execution_instance.scheduler
+
+    locality_hint = "test-locality"
+
+    # Queue job with locality hint
+    bot_name, bot_id, _, job_name = mock_queue_job_action(scheduler, locality_hint=locality_hint)
+
+    # Complete the job with exit_code=1 (failure)
+    leases = scheduler.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, [])
+    assert len(leases) == 1
+    lease = leases[0]
+    lease.state = LeaseState.ACTIVE.value
+    leases = scheduler.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, [lease])
+    assert len(leases) == 1
+    lease = leases[0]
+
+    # Complete with failed exit code
+    test_action_result = remote_execution_pb2.ActionResult(exit_code=1, stdout_raw=b"failed")
+    lease.result.Pack(test_action_result)
+    lease.state = LeaseState.COMPLETED.value
+    scheduler.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, [lease])
+
+    # Verify locality hint was NOT recorded due to non-zero exit code
+    with scheduler._sql.session() as session:
+        locality_hints = (
+            session.execute(select(BotLocalityHintEntry).where(BotLocalityHintEntry.bot_name == bot_name))
+            .scalars()
+            .all()
+        )
+        assert len(locality_hints) == 0
+
+
+def test_proactive_fetch_to_capacity_enabled(controller_postgres):
+    """Test that with proactive_fetch_to_capacity=True, bot capacity is used for proactive fetching."""
+    scheduler: Scheduler = controller_postgres.execution_instance.scheduler
+    scheduler.proactive_fetch_to_capacity = True
+
+    # Create a bot with capacity 3
+    bot_name, bot_id, _, _ = mock_queue_job_action(scheduler, bot_capacity=3, do_not_cache=True, assign=False)
+    # Add another 3 jobs
+    for _ in range(3):
+        mock_queue_job_action(scheduler, create_bot=False, do_not_cache=True, assign=False)
+
+    # Bot should have 3 active leases (full capacity)
+    leases = scheduler.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, [])
+    assert len(leases) == 3
+
+    # Complete only 1 job
+    completed_leases = [leases[0]]
+    completed_leases[0].state = LeaseState.COMPLETED.value
+
+    # With proactive_fetch_to_capacity=True, should fetch up to full capacity (3 jobs total)
+    remaining_leases = leases[1:]  # The 2 jobs still active
+    new_leases = scheduler.synchronize_bot_leases(
+        bot_name, bot_id, BotStatus.OK.value, completed_leases + remaining_leases
+    )
+
+    # Should have 3 total leases: 2 remaining + 1 newly fetched to reach full capacity
+    assert len(new_leases) == 3
+
+
 def test_bot_proactively_fetch(controller_postgres):
     """Test that bots can proactively fetch jobs when bot_proactively_fetch is enabled."""
     scheduler: Scheduler = controller_postgres.execution_instance.scheduler
@@ -1196,8 +1348,43 @@ def test_bot_proactively_fetch(controller_postgres):
         assert jobs[1].worker_name == bot_id
 
 
-def test_schedule_by_instances(controller_postgres):
+def test_bot_proactively_fetch_near_max_quota(controller_postgres):
+    scheduler: Scheduler = controller_postgres.execution_instance.scheduler
+    # Create quota configs for two instances, allowing only 1 concurrent job each
+    scheduler.put_instance_quota("sql", "linux", 1, 1)
+    scheduler.put_instance_quota("other", "linux", 1, 1)
 
+    # Enqueue 2 jobs
+    bot_name, bot_id, _, _ = mock_queue_job_action(scheduler, do_not_cache=True, skip_cache_lookup=True, assign=False)
+    _ = mock_queue_job_action(scheduler, do_not_cache=True, skip_cache_lookup=True, assign=False)
+    scheduler.assign_job_by_priority()  # successful
+    scheduler.assign_job_by_priority(failure_backoff=60)  # failure, adding `schedule_after`
+
+    leases = scheduler.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, [])
+    assert len(leases) == 1
+    leases[0].state = LeaseState.COMPLETED.value
+    leases = scheduler.synchronize_bot_leases(bot_name, bot_id, BotStatus.OK.value, leases)
+    assert len(leases)
+
+    with scheduler._sql.session() as session:
+        jobs = session.execute(select(JobEntry)).scalars().all()
+        assert len(jobs) == 2
+        assert jobs[0].stage == OperationStage.COMPLETED.value
+        # Assert that the second job is scheduled
+        assert jobs[1].assigned
+        assert jobs[1].worker_name == bot_id
+
+        # Assert that the bot has no more capacity to fetch more jobs
+        bot_entry = session.execute(select(BotEntry).where(BotEntry.name == bot_name)).scalar_one()
+        assert bot_entry.capacity == 0
+
+    # Check usage is recorded correctly
+    usage = scheduler.get_instance_quota("sql", "linux")
+    assert usage is not None
+    assert usage.current_usage == 1
+
+
+def test_schedule_by_instances(controller_postgres):
     scheduler: Scheduler = controller_postgres.execution_instance.scheduler
 
     # Enqueue a job without assigning
@@ -1219,7 +1406,6 @@ def test_schedule_by_instances(controller_postgres):
 
 
 def test_schedule_by_instances_isolation(controller_postgres):
-
     scheduler: Scheduler = controller_postgres.execution_instance.scheduler
 
     # Enqueue a job without assigning
@@ -1667,6 +1853,10 @@ def test_assign_job_by_cohort_preemption(controller_postgres: ExecutionControlle
             b_jobs = session.execute(select(JobEntry).where(JobEntry.instance_name == "b")).scalars().all()
             assert len([job for job in b_jobs if job.assigned]) == 1
 
+            capacity = session.execute(select(BotEntry.capacity)).scalar_one_or_none()
+            assert capacity is not None
+            assert capacity == 0
+
     # Check the usages are right
     a_usage = scheduler.get_instance_quota("a", "linux")
     b_usage = scheduler.get_instance_quota("b", "linux")
